from typing import Callable, Literal, TypeVar, Union, overload

from ._protocols import IActionCallback, IActionsCallback

T = TypeVar("T")
Decorator = Callable[[T], T]


@overload
def setup(func: IActionCallback) -> IActionCallback:
    ...


@overload
def setup(*, scope: Literal["action"] = "action") -> Decorator[IActionCallback]:
    ...


@overload
def setup(*, scope: Literal["session"]) -> Decorator[IActionsCallback]:
    ...


def setup(
    *args, **kwargs
) -> Union[IActionCallback, Decorator[IActionCallback], Decorator[IActionsCallback]]:
    """Run code before any actions start, or before each separate action.

    Receives as an argument the action or actions that will be run.

    Can be used as a decorator without arguments:

    ```python
    from robocorp.actions import setup

    @setup
    def my_fixture(action):
        print(f"Before action: {action.name}")
    ```

    Alternatively, can be called with a `scope` argument to decide when
    the fixture is run:

    ```python
    from robocorp.actions import setup

    @setup(scope="action")
    def before_each(action):
        print(f"Running action '{action.name}'")

    @setup(scope="session")
    def before_all(actions):
        print(f"Running {len(actions)} action(s)")
    ```

    By default, runs setups in `action` scope.

    The `setup` fixture also allows running code after the execution,
    if it `yield`s the execution to the action(s):

    ```python
    import time
    from robocorp.actions import setup

    @setup
    def measure_time(action):
        start = time.time()
        yield  # Action executes here
        duration = time.time() - start
        print(f"Action took {duration} seconds")

    @action
    def my_long_action():
        ...
    ```

    **Note:** If fixtures are defined in another file, they need to be imported
     in the main actions file to be taken into use
    """
    from robocorp.tasks import setup as _setup

    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        return _setup(*args, **kwargs)

    scope = kwargs.pop("scope", "action")
    if scope == "action":
        scope = "task"

    elif scope == "session":
        pass

    else:
        raise ValueError(f"Unknown scope '{scope}', expected 'action' or 'session'")

    kwargs["scope"] = scope
    return _setup(*args, **kwargs)


@overload
def teardown(func: IActionCallback) -> IActionCallback:
    ...


@overload
def teardown(*, scope: Literal["action"] = "action") -> Decorator[IActionCallback]:
    ...


@overload
def teardown(*, scope: Literal["session"]) -> Decorator[IActionsCallback]:
    ...


def teardown(
    *args, **kwargs
) -> Union[IActionCallback, Decorator[IActionCallback], Decorator[IActionsCallback]]:
    """Run code after actions have been run, or after each separate action.

    Receives as an argument the action or actions that were executed, which
    contain (among other things) the resulting status and possible error message.

    Can be used as a decorator without arguments:

    ```python
    from robocorp.actions import teardown

    @teardown
    def my_fixture(action):
        print(f"After action: {action.name})
    ```

    Alternatively, can be called with a `scope` argument to decide when
    the fixture is run:

    ```python
    from robocorp.actions import teardown

    @teardown(scope="action")
    def after_each(action):
        print(f"Action '{action.name}' status is '{action.status}'")

    @teardown(scope="session")
    def after_all(actions):
        print(f"Executed {len(actions)} action(s)")
    ```

    By default, runs teardowns in `action` scope.

    **Note:** If fixtures are defined in another file, they need to be imported
     in the main actions file to be taken into use
    """
    from robocorp.tasks import teardown as _teardown

    if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
        return _teardown(*args, **kwargs)

    scope = kwargs.pop("scope", "action")
    if scope == "action":
        scope = "task"

    elif scope == "session":
        pass

    else:
        raise ValueError(f"Unknown scope '{scope}', expected 'action' or 'session'")

    kwargs["scope"] = scope
    return _teardown(*args, **kwargs)



import ast as ast_module
from dataclasses import dataclass
from typing import Any, Iterator, List, Optional, Tuple, TypedDict, Union, overload


def _iter_nodes(
    node, internal_stack: Optional[List[Any]] = None, recursive=True
) -> Iterator[Tuple[List[Any], Any]]:
    """
    :note: the yielded stack is actually always the same (mutable) list, so,
    clients that want to return it somewhere else should create a copy.
    """
    stack: List[Any]
    if internal_stack is None:
        stack = []
        if node.__class__.__name__ != "File":
            stack.append(node)
    else:
        stack = internal_stack

    if recursive:
        for _field, value in ast_module.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast_module.AST):
                        yield stack, item
                        stack.append(item)
                        yield from _iter_nodes(item, stack, recursive=True)
                        stack.pop()

            elif isinstance(value, ast_module.AST):
                yield stack, value
                stack.append(value)

                yield from _iter_nodes(value, stack, recursive=True)

                stack.pop()
    else:
        # Not recursive
        for _field, value in ast_module.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast_module.AST):
                        yield stack, item

            elif isinstance(value, ast_module.AST):
                yield stack, value


class PositionTypedDict(TypedDict):
    # Line position in a document (zero-based).
    line: int

    # Character offset on a line in a document (zero-based). Assuming that
    # the line is represented as a string, the `character` value represents
    # the gap between the `character` and `character + 1`.
    #
    # If the character value is greater than the line length it defaults back
    # to the line length.
    character: int


class RangeTypedDict(TypedDict):
    start: PositionTypedDict
    end: PositionTypedDict


class DiagnosticsTypedDict(TypedDict, total=False):
    # The range at which the message applies.
    range: RangeTypedDict

    # The diagnostic's severity. Can be omitted. If omitted it is up to the
    # client to interpret diagnostics as error, warning, info or hint.
    severity: Optional[int]  # DiagnosticSeverity

    # The diagnostic's code, which might appear in the user interface.
    code: Union[int, str]

    # An optional property to describe the error code.
    #
    # @since 3.16.0
    codeDescription: Any

    # A human-readable string describing the source of this
    # diagnostic, e.g. 'typescript' or 'super lint'.
    source: Optional[str]

    # The diagnostic's message.
    message: str

    # Additional metadata about the diagnostic.
    #
    # @since 3.15.0
    tags: list  # DiagnosticTag[];

    # An array of related diagnostic information, e.g. when symbol-names within
    # a scope collide all definitions can be marked via this property.
    relatedInformation: list  # DiagnosticRelatedInformation[];

    # A data entry field that is preserved between a
    # `textDocument/publishDiagnostics` notification and
    # `textDocument/codeAction` request.
    #
    # @since 3.16.0
    data: Optional[Any]  # unknown;


class DiagnosticSeverity:
    Error = 1
    Warning = 2
    Information = 3
    Hint = 4


class Error(object):
    __slots__ = "msg start end severity tags data".split(" ")

    def __init__(
        self,
        msg: str,
        start: Tuple[int, int],
        end: Tuple[int, int],
        severity: int = DiagnosticSeverity.Error,
    ):
        """
        Note: `start` and `end` are tuples with (line, col).
        """
        self.msg = msg
        self.start = start
        self.end = end
        self.severity = severity
        self.data: Any = None

    def to_dict(self):
        ret = {
            "msg": self.msg,
            "start": self.start,
            "end": self.end,
            "severity": self.severity,
        }
        tags = getattr(self, "tags", None)
        if tags:
            ret["tags"] = tags
        if self.data is not None:
            ret["data"] = self.data
        return ret

    def __repr__(self):
        import json

        return json.dumps(self.to_dict())

    __str__ = __repr__

    def to_lsp_diagnostic(self) -> DiagnosticsTypedDict:
        tags = getattr(self, "tags", None)
        ret: DiagnosticsTypedDict = {
            "range": {
                "start": {"line": self.start[0], "character": self.start[1]},
                "end": {"line": self.end[0], "character": self.end[1]},
            },
            "severity": self.severity,
            "source": "robocorp.actions",
            "message": self.msg,
        }
        if tags:
            ret["tags"] = tags
        if self.data is not None:
            ret["data"] = self.data
        return ret


def _make_error(
    node: ast_module.AST, msg, severity=DiagnosticSeverity.Error, coldelta=0
) -> Error:
    if isinstance(node, ast_module.FunctionDef):
        return Error(
            msg,
            (node.lineno, node.col_offset + coldelta),
            (node.lineno, node.col_offset + coldelta + len(node.name)),
            severity,
        )
    else:
        return Error(
            msg,
            (node.lineno, node.col_offset + coldelta),
            (
                node.end_lineno or node.lineno,
                (node.end_col_offset or node.col_offset) + coldelta,
            ),
            severity,
        )


def _check_return_type(node: ast_module.FunctionDef) -> Iterator[Error]:
    return_type = node.returns
    if not return_type:
        yield _make_error(
            node,
            "@action without return annotation defined "
            "(`str` considered as the return type).",
            severity=DiagnosticSeverity.Warning,
            coldelta=4,
        )


def _check_return_statement(node: ast_module.FunctionDef) -> Iterator[Error]:
    for _stack, maybe_return_node in _iter_nodes(node, recursive=True):
        if isinstance(maybe_return_node, ast_module.Return):
            break
    else:
        # Return not found!
        yield _make_error(
            node,
            "@action without a `return` (an `@action` must have a return value).",
            coldelta=4,
        )


def _check_arguments(node: ast_module.FunctionDef, docstring: str) -> Iterator[Error]:
    arguments = node.args
    if arguments.args:
        param_name_to_description = {}
        if docstring:
            import docstring_parser

            contents = docstring_parser.parse(docstring)
            for docparam in contents.params:
                if docparam.description:
                    param_name_to_description[docparam.arg_name] = docparam.description

            if (
                not contents.short_description
                or contents.short_description.strip() == "Args:"
            ):
                yield _make_error(
                    node,
                    "No description found in docstring. "
                    "Please update docstring to add it as an LLM needs this info "
                    "to understand when to call this action.",
                    coldelta=4,
                )
                return

        for arg in arguments.args:
            desc = param_name_to_description.pop(arg.arg, None)
            if not desc:
                yield _make_error(
                    arg,
                    f"Parameter: `{arg.arg}` documentation not found in docstring. "
                    "Please update docstring to add it as an LLM needs this info "
                    "to understand what needs to be passed to call this action.",
                )

            if not arg.annotation and not arg.type_comment:
                yield _make_error(
                    arg,
                    f"Parameter: `{arg.arg}` without annotation "
                    "(considering `str` as the expected type).",
                    severity=DiagnosticSeverity.Warning,
                )


def iter_lint_errors(action_contents_file: str | bytes) -> Iterator[Error]:
    ast = ast_module.parse(action_contents_file, "<string>")
    for _stack, node in _iter_nodes(ast, recursive=False):
        if isinstance(node, ast_module.FunctionDef):
            for decorator in node.decorator_list:
                if isinstance(decorator, ast_module.Name) and decorator.id == "action":
                    # We found an @action. Do the needed checks.

                    # Check for docstring
                    docstring: str = ast_module.get_docstring(node) or ""
                    if docstring:
                        docstring = docstring.strip()

                    if not docstring:
                        yield _make_error(
                            node,
                            "@action docstring not found. Please define it as this is what "
                            "an LLM would use to decide whether to call this action.",
                            coldelta=4,
                        )

                    yield from _check_return_type(node)
                    yield from _check_return_statement(node)
                    yield from _check_arguments(node, docstring)


_severity_id_to_severity = {
    DiagnosticSeverity.Error: "error",
    DiagnosticSeverity.Warning: "warning",
    DiagnosticSeverity.Information: "information",
    DiagnosticSeverity.Hint: "hint",
}


@dataclass
class FormattedLintResult:
    message: str
    found_critical: bool


class LintResultTypedDict(TypedDict):
    file: str
    errors: List[DiagnosticsTypedDict]


@overload
def format_lint_results(
    lint_result: LintResultTypedDict,
) -> Optional[FormattedLintResult]:
    ...


@overload
def format_lint_results(
    lint_result: dict,
) -> Optional[FormattedLintResult]:
    ...


def format_lint_results(
    lint_result: Union[dict, LintResultTypedDict],
) -> Optional[FormattedLintResult]:
    f = lint_result.get("file")
    errors = lint_result.get("errors")
    if not isinstance(f, str) or not isinstance(errors, list):
        return None

    msg_parts = []
    msg_parts.append(f"Action lint error(s) found at file: {f}")
    msg_parts.append(
        "(note: it's possible to skip linting by passing `--skip-lint` in the command line)"
    )
    found_critical = False
    for error in errors:
        pos = error.get("range", {}).get("start", {})
        line = pos.get("line", -1)

        severity_level = error.get("severity", -1)
        if severity_level == DiagnosticSeverity.Error:
            found_critical = True

        severity_msg = _severity_id_to_severity.get(
            severity_level if severity_level is not None else -1, "Unknown severity"
        )
        message = error.get("message")
        msg_parts.append(f"{severity_msg.title()} (line {line}): {message}")

    return FormattedLintResult(
        message="\n".join(msg_parts), found_critical=found_critical
    )



from dataclasses import asdict
from pathlib import Path
from typing import Callable, Optional, overload

from ._action_options import ActionOptions
from ._fixtures import setup, teardown
from ._protocols import IAction, Status

__version__ = "0.0.7"
version_info = [int(x) for x in __version__.split(".")]


@overload
def action(func: Callable) -> Callable:
    ...


@overload
def action(func: Callable, **kwargs: Optional[ActionOptions]) -> Callable:
    ...


def action(*args, **kwargs):
    """
    Decorator for actions (entry points) which can be executed by `robocorp.actions`.

    i.e.:

    If a file such as actions.py has the contents below:

    ```python
    from robocorp.actions import action

    @action
    def enter_user() -> str:
        ...
    ```

    It'll be executable by robocorp actions as:

    python -m robocorp.actions run actions.py -a enter_user

    Args:
        func: A function which is a action to `robocorp.actions`.
        is_consequential: Whether the action is consequential or not. This will add `x-openai-isConsequential: true` to the action metadata and shown in OpenApi spec.
    """

    def decorator(*args, **kwargs):
        # i.e.: This is just a thin layer for the task decorator at this point
        # (it may be extended in the future...).
        from robocorp.tasks import task

        options = ActionOptions(**kwargs)

        return task(*args, **asdict(options))

    if args and callable(args[0]):
        return decorator(args[0], **kwargs)

    return lambda func: decorator(func, **kwargs)


def session_cache(func):
    """
    Provides decorator which caches return and clears automatically when all
    actions have been run.

    A decorator which automatically cache the result of the given function and
    will return it on any new invocation until robocorp-actions finishes running
    all actions.

    The function may be either a generator with a single yield (so, the first
    yielded value will be returned and when the cache is released the generator
    will be resumed) or a function returning some value.

    Args:
        func: wrapped function.
    """
    from robocorp.tasks import session_cache

    return session_cache(func)


def action_cache(func):
    """
    Provides decorator which caches return and clears it automatically when the
    current action has been run.

    A decorator which automatically cache the result of the given function and
    will return it on any new invocation until robocorp-actions finishes running
    the current action.

    The function may be either a generator with a single yield (so, the first
    yielded value will be returned and when the cache is released the generator
    will be resumed) or a function returning some value.

    Args:
        func: wrapped function.
    """
    from robocorp.tasks import task_cache

    return task_cache(func)


def get_output_dir() -> Optional[Path]:
    """
    Provide the output directory being used for the run or None if there's no
    output dir configured.
    """
    from robocorp.tasks import get_output_dir

    return get_output_dir()


def get_current_action() -> Optional[IAction]:
    """
    Provides the action which is being currently run or None if not currently
    running an action.
    """
    from robocorp.tasks import get_current_task

    return get_current_task()


__all__ = [
    "action",
    "setup",
    "teardown",
    "session_cache",
    "action_cache",
    "get_output_dir",
    "get_current_action",
    "IAction",
    "Status",
]



import sys
import typing
from contextlib import contextmanager

from robocorp.tasks._argdispatch import _ArgDispatcher


def _translate(msg):
    return (
        msg.replace("task", "action")
        .replace("Task", "Action")
        .replace("TASK", "ACTION")
    )


class _ActionsArgDispatcher(_ArgDispatcher):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register("list")(self._list)
        self.register("run")(self._run)

    @contextmanager
    def _register_lint(self, stream: typing.IO):
        from pathlib import Path

        from robocorp.tasks import _hooks

        from robocorp.actions import _lint_action
        from robocorp.actions._lint_action import format_lint_results

        files_found = set()

        def on_func_found(func, *args, **kwargs) -> None:
            import json

            from robocorp.actions._lint_action import (
                DiagnosticSeverity,
                DiagnosticsTypedDict,
                LintResultTypedDict,
            )

            filename = Path(func.__code__.co_filename).absolute()

            if filename not in files_found:
                files_found.add(filename)
                errors = list(
                    x.to_lsp_diagnostic()
                    for x in _lint_action.iter_lint_errors(filename.read_bytes())
                )
                if errors:
                    found_critical = False
                    error: DiagnosticsTypedDict
                    lint_result_contents: LintResultTypedDict = {
                        "file": str(filename),
                        "errors": errors,
                    }
                    lint_result = {"lint_result": lint_result_contents}

                    for error in errors:
                        if error["severity"] == DiagnosticSeverity.Error:
                            found_critical = True

                    if found_critical:
                        stream.write(json.dumps(lint_result))
                        stream.flush()
                        sys.exit(1)

                    # Critical not found: print to stderr
                    formatted = format_lint_results(lint_result_contents)
                    if formatted is not None:
                        sys.stderr.write(formatted.message)
                        sys.stderr.flush()

        with _hooks.on_task_func_found.register(on_func_found):
            yield

    def _list(self, *args, **kwargs):
        from contextlib import nullcontext, redirect_stdout

        from robocorp.tasks import _commands

        skip_lint = kwargs.pop("skip_lint", True)

        original_stdout = sys.stdout

        with redirect_stdout(sys.stderr):
            ctx = (
                self._register_lint(original_stdout) if not skip_lint else nullcontext()
            )
            with ctx:
                return _commands.list_tasks(*args, __stream__=original_stdout, **kwargs)

    def _run(self, *args, **kwargs):
        from robocorp.tasks import _commands

        return _commands.run(*args, **kwargs)

    def _get_description(self):
        return "Robocorp Actions library"

    def _add_task_argument(self, run_parser):
        run_parser.add_argument(
            "-a",
            "--action",
            dest="task_name",
            help="The name of the action that should be run.",
            action="append",
        )

    def _add_lint_argument(self, parser):
        parser.add_argument(
            "--skip-lint",
            dest="skip_lint",
            action="store_true",
            default=False,
            help="Skip `@action` linting when an action is found (by default any "
            "`@action` is linted for errors when found).",
        )

    def _create_run_parser(self, main_parser):
        parser = super()._create_run_parser(main_parser)
        # Not adding to run at this point.
        # self._add_lint_argument(parser)
        return parser

    def _create_list_tasks_parser(self, main_parser):
        parser = super()._create_list_tasks_parser(main_parser)
        self._add_lint_argument(parser)
        return parser

    def _get_argument_parser_class(self):
        import argparse

        class ArgumentParserTranslated(argparse.ArgumentParser):
            def format_usage(self):
                return _translate(argparse.ArgumentParser.format_usage(self))

            def format_help(self):
                return _translate(argparse.ArgumentParser.format_help(self))

        return ArgumentParserTranslated

    def _dispatch(self, parsed) -> int:
        # Custom dispatch as we need to account for custom flags.
        if not parsed.command:
            self._create_argparser().print_help()
            return 1

        method = self._name_to_func[parsed.command]
        dct = parsed.__dict__.copy()
        dct.pop("command")

        return method(**dct)



import sys


def main(args=None, exit: bool = True) -> int:
    """Entry point for running actions from robocorp-actions."""
    from robocorp.actions._args_dispatcher import _ActionsArgDispatcher

    # Note: robocorp-actions is almost only robocorp-tasks but with a few
    # different APIs.
    #
    # As both are under our control, we rely on internal APIs to make
    # it seem to the external world as everything is action-based (and not
    # task-based).

    if args is None:
        args = sys.argv[1:]

    from robocorp.tasks import _constants
    from robocorp.tasks.cli import main

    _constants.DEFAULT_TASK_SEARCH_GLOB = "*action*.py"
    _constants.MODULE_ENTRY_POINT = "robocorp.actions"

    return main(args, exit, argument_dispatcher=_ActionsArgDispatcher())
